@@ -40,5 +40,6 @@ DROP TABLE IF EXISTS authors;
 CREATE TABLE authors (
     id INT NOT NULL AUTO_INCREMENT PRIMARY KEY,
     email VARCHAR(100) NOT NULL UNIQUE,
-    name VARCHAR(100) NOT NULL
+    name VARCHAR(100) NOT NULL,
+    hashed_password VARCHAR(100) NOT NULL
 );
